/*
    Fintek F81504A / F81508A / F81512A
*/

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/tty.h>
#include <linux/ratelimit.h>
#include <linux/tty_flip.h>

#include <linux/serial_reg.h>
#include <linux/serial_core.h>
#include <linux/8250_pci.h>
#include <linux/bitops.h>
#include <linux/serial_8250.h>
#include <linux/version.h>
#include <linux/delay.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <linux/mutex.h>

#include <linux/io.h>
#include <linux/uaccess.h>

#include <linux/serial_8250.h>
#include <linux/serial_reg.h>
#include <linux/dmaengine.h>
#include <linux/gpio.h>
#include <linux/pm_runtime.h>

#define VERSION "v1.04-20241025"

#ifndef ACCESS_ONCE
#define __ACCESS_ONCE(x)                                               \
	({                                                             \
		__maybe_unused typeof(x) __var = (__force typeof(x))0; \
		(volatile typeof(x) *)&(x);                            \
	})
#define ACCESS_ONCE(x) (*__ACCESS_ONCE(x))
#endif

#define F81504A_INTERNAL_LOOP 0
#define F81504A_CLK_DIV_2 0

#define EXTRA_GPIO_CHANGE_MODE 1
#define USER_MODE 0
#define NO_GPIO_DEV 0xff
#define USER_FIX_MODE 0xff
#define F81504_I2C_CONF 0xa00

#define F81504A_MAX_SERIAL 12
#define MAX_FINTEK_SERIAL_PORT F81504A_MAX_SERIAL
#define F81504A_DEF_BAUDRATE 9600

#define F81504A_UART_CONF_ACCESS_BASE 0xAA0
#define F81504A_UART_CONF(x) (F81504A_UART_CONF_ACCESS_BASE + 8 * x)

#define F81504A_UART_BASE_ADDR 0x2A0
#define F81504A_UART_CTRL_REG(x) (F81504A_UART_BASE_ADDR + 0 + (8 * x))
#define F81504A_UART_FIFO_REG(x) (F81504A_UART_BASE_ADDR + 1 + (8 * x))
#define F81504A_UART_9BIT_ADDR_REG(x) (F81504A_UART_BASE_ADDR + 2 + (8 * x))
#define F81504A_UART_9BIT_MASK_REG(x) (F81504A_UART_BASE_ADDR + 3 + (8 * x))
#define F81504A_UART_BASE_MSB_REG(x) (F81504A_UART_BASE_ADDR + 4 + (8 * x))
#define F81504A_UART_BASE_LSB_REG(x) (F81504A_UART_BASE_ADDR + 5 + (8 * x))
#define F81504A_UART_IRQ_CHAN_REG(x) (F81504A_UART_BASE_ADDR + 6 + (8 * x))
#define F81504A_UART_IRQ_SHARE_REG(x) (F81504A_UART_BASE_ADDR + 7 + (8 * x))

/*
 * Control Register (offset +0h)
 *
 * Bit7~6:	Clock select
 *			00: 1.8432MHz
 *			01: 18.432MHz
 *			10: 24MHz
 *			11: 14.769MHz
 *
 * Bit0:	UART enable
 */
#define F81504A_UART_EN BIT(0)
#define F81504A_CLK_1_846_MHZ 0
#define F81504A_CLK_18_46_MHZ BIT(6)
#define F81504A_CLK_24_MHZ BIT(7)
#define F81504A_CLK_48_MHZ (BIT(7) | BIT(5))
#define F81504A_CLK_14_77_MHZ (BIT(7) | BIT(6))
#define F81504A_CLK_MASK GENMASK(7, 5)

/*
 * FIFO Select Register (offset +1h)
 *
 * Bit5~4:	Trigger level multiple
 *			00: 1x
 *			01: 2x
 *			10: 4x
 *			11: 8x
 *
 * Bit1~0:	FIFO size
 *			00: 16 bytes FIFO
 *			01: 32 bytes FIFO
 *			10: 64 bytes FIFO
 *			11: 128 bytes FIFO
 */
#define F81504A_TRIGGER_8X_MODE (BIT(5) | BIT(4))
#define F81504A_FIFO_256_MODE (BIT(2) | BIT(1) | BIT(0))

/*
 * IRQ Share Register (offset +7h)
 *
 * Bit5:	Invert direction(RTS) when Bit4 enabled (RTS pin high when TX)
 * Bit4:	Auto direction(RTS) control (RTS pin Low when TX)
 * Bit0:	IRQ Share enable
 */
#define F81504A_UART_RS485_INV BIT(5)
#define F81504A_UART_RS485_ENABLE BIT(4)
#define F81504A_UART_IRQ_SHARE BIT(0)

typedef enum {
	transceiver_type_f81439,
	transceiver_type_f81435,
	transceiver_type_f81433,
	transceiver_type_invalid,
} transceiver_type;

struct f81504a_priv {
	struct pci_device_id dev_id;

	struct serial_rs485 rs485[F81504A_MAX_SERIAL];
	int line_data[F81504A_MAX_SERIAL];

#if EXTRA_GPIO_CHANGE_MODE
	bool is_i2c_mem;
	unsigned char __iomem *i2c_mem;
	int gpio_dev_id;
	int gpio_mode[MAX_FINTEK_SERIAL_PORT];
	transceiver_type transceiver_type[MAX_FINTEK_SERIAL_PORT];
#endif

	struct mutex locker;
	int irq_cnt;
};

static bool force_high_speed;
module_param(force_high_speed, bool, S_IRUGO);
MODULE_PARM_DESC(force_high_speed, "Force baudrate to 6Mbit");

/*
 0: memory
 1: io
 2: configure
 */

static unsigned int mmio_mode = 0;
module_param(mmio_mode, uint, S_IRUGO);
MODULE_PARM_DESC(mmio_mode, "0: mem, 1: io, 2: conf");

static unsigned int loopback = 0;
module_param(loopback, uint, S_IRUGO);
MODULE_PARM_DESC(loopback, "loopback");

static unsigned int en_msi = 1;
module_param(en_msi, uint, S_IRUGO);
MODULE_PARM_DESC(
	en_msi, "Enable Message Signal Interrupt(MSI) mode, (default enable)");

static unsigned int max_msi = 16;
module_param(max_msi, uint, S_IRUGO);
MODULE_PARM_DESC(max_msi, "Max MSI interrupt (default 16)");

#if F81504A_CLK_DIV_2
static u32 const baudrate_table[] = { 115200, 921600, 1152000, 1500000,
				      3000000 };
#else
static u32 const baudrate_table[] = { 115200 * 2, 921600 * 2, 1152000 * 2,
				      1500000 * 2, 3000000 * 2 };
#endif

static u8 const clock_table[] = { F81504A_CLK_1_846_MHZ, F81504A_CLK_14_77_MHZ,
				  F81504A_CLK_18_46_MHZ, F81504A_CLK_24_MHZ,
				  F81504A_CLK_48_MHZ };

static spinlock_t global_lock;

struct uart_8250_port *serial8250_get_port(int line);

#if 0
static unsigned int f81504a_conf_serial_fixup_in(struct uart_port *port,
						int offset)
{
	struct pci_dev *pdev = to_pci_dev(port->dev);
	u8 port_idx = *(u8 *)port->private_data;
	unsigned int tmp;

	pci_read_config_dword(pdev, F81504A_UART_CONF(port_idx) + offset, &tmp);

	return tmp;
}
#endif

static unsigned int f81504a_conf_serial_in(struct uart_port *port, int offset)
{
	struct pci_dev *pdev = to_pci_dev(port->dev);
	u8 port_idx = *(u8 *)port->private_data;
	u8 tmp;

	pci_read_config_byte(pdev, F81504A_UART_CONF(port_idx) + offset, &tmp);

#if 0
	if (port_idx == 0) {
		pr_info("conf read: %x, data: %x\n",
			F81504A_UART_CONF(port_idx) + offset, tmp);
	}
#endif
	return tmp;
}

static void f81504a_conf_serial_out(struct uart_port *port, int offset,
				    int data)
{
	struct pci_dev *pdev = to_pci_dev(port->dev);
	u8 idx = *(u8 *)port->private_data;

#if 0
	struct f81504a_priv *priv = pci_get_drvdata(pdev);
	u8 *ptr;
	u8 lcr;

	if (offset == 0) {
		lcr = serial_port_in(port, UART_LCR);
		if (lcr & UART_LCR_DLAB) {
			pci_write_config_byte(pdev, F81504A_UART_CONF(idx) + offset, data);
			return;
		}

		/* DMA TX data */
		pci_write_config_dword(pdev, F81504A_DMA_TX_ADDR + idx * 4,
					(u32) priv->tx[idx]);
		
		ptr = (u8*) priv->dma_addr_tx;
		ACCESS_ONCE(*ptr) = data;

		pci_write_config_byte(pdev, F81504A_DMA_TX_CNT + idx, 0);
	}
#else
	pci_write_config_byte(pdev, F81504A_UART_CONF(idx) + offset, data);
#endif
}

#if EXTRA_GPIO_CHANGE_MODE
const static int m_accpet_mode_f81439[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };
const static int m_accpet_mode_f81435[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };
const static int m_accpet_mode_f81433[] = { 1, 2, 3, 4, 5, 7, -1 };

const static int m_rs485_f81439[] = { 0, 0, 3, 1, 0, 0, 3, 0, -1 };
const static int m_rs485_f81435[] = { 1, 0, 1, 0, 3, 3, 3, 3, -1 };
const static int m_rs485_f81433[] = { 0, 0, 0, 0, 3, 3, 0, 0, -1 };

const static int *m_nAcceptArray[transceiver_type_invalid] = {
	[transceiver_type_f81433] = m_accpet_mode_f81433,
	[transceiver_type_f81435] = m_accpet_mode_f81435,
	[transceiver_type_f81439] = m_accpet_mode_f81439,
};

const static int *m_n_rs485_Array[transceiver_type_invalid] = {
	[transceiver_type_f81433] = m_rs485_f81433,
	[transceiver_type_f81435] = m_rs485_f81435,
	[transceiver_type_f81439] = m_rs485_f81439,
};

static int check_transceiver_mode(transceiver_type type, int mode)
{
	const int *tb;
	int i;

	if (type < transceiver_type_f81439 || type >= transceiver_type_invalid)
		return -1;

	tb = m_nAcceptArray[type];

	for (i = 0; tb[i] != -1; ++i) {
		if (mode == tb[i])
			return 0;
	}

	return -EINVAL;
}

static void writeb_delayed(u8 data, u8 __iomem *mem)
{
	writeb(data, mem);
	readb(mem);
}

static int __WriteI2CData_F81504_MEM__(struct pci_dev *pdev, unsigned int dev,
				       unsigned int addr, unsigned int data)
{
	struct f81504a_priv *priv = pci_get_drvdata(pdev);
	int count;
	u8 tmp;

	/* Reset */
	writeb_delayed(0x80, priv->i2c_mem + 0x07);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x07], 0x80);

	count = 1000;
	while (--count) {
		tmp = readb(priv->i2c_mem + 0x07);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	writeb_delayed(dev & 0xff, priv->i2c_mem + 0x02);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x02], dev & 0xff);

	/* enable WR_CMD_EN */
	writeb_delayed(0x03, priv->i2c_mem + 0x06);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x06], 0x03);

	/* clear busy flag */
	writeb_delayed(0x17, priv->i2c_mem + 0x01);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x01], 0x17);

	/* Set Reg addr */
	writeb_delayed(addr & 0xff, priv->i2c_mem + 0x03);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x03], addr & 0xff);

	/* Set Data */
	writeb_delayed(data & 0xff, priv->i2c_mem + 0x04);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x04], data & 0xff);

	/* Start Read, 0x02: EEPROM Write, 0x08: DBI Write?? */
	writeb_delayed(0x02, priv->i2c_mem + 0x00);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x00], 0x02);

	/* check not busy */
	count = 100;
	while (!(readb(priv->i2c_mem + 0x01) & 0x01) && count) {
		count -= 1;
		udelay(100);
	}

	/* check error */
	if (readb(priv->i2c_mem + 0x01) & (BIT(1) | BIT(2))) {
#if 0		
		if (readb(priv->i2c_mem + 0x01) & (BIT(1)))
			pr_info("%s: Write I2C NACK\n", __func__);

		if (readb(priv->i2c_mem + 0x01) & (BIT(2)))
			pr_info("%s: Write I2C TIMEOUT\n", __func__);
#endif
		return -ETIMEDOUT;
	}

	return 0;
}

static int __ReadI2CData_F81504_MEM__(struct pci_dev *pdev, unsigned int dev,
				      unsigned int addr, unsigned int *data)
{
	struct f81504a_priv *priv = pci_get_drvdata(pdev);
	int count;
	u8 tmp;

	/* Reset */
	writeb_delayed(0x80, priv->i2c_mem + 0x07);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x07], 0x80);

	count = 1000;
	while (--count) {
		tmp = readb(priv->i2c_mem + 0x07);
		if (!tmp)
			break;
	}

	if (!count) {
		//fprintf(stderr, "%s: reset I2C TIMEOUT\n", __func__);
		return -ETIMEDOUT;
	}

	/* device address */
	writeb_delayed(dev & 0xff, priv->i2c_mem + 0x02);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x02], dev & 0xff);

	/* enable WR_CMD_EN */
	writeb_delayed(0x03, priv->i2c_mem + 0x06);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x06], 0x03);

	/* clear busy flag */
	writeb_delayed(0x17, priv->i2c_mem + 0x01);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x01], 0x17);

	/* Set Reg addr */
	writeb_delayed(addr & 0xff, priv->i2c_mem + 0x03);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x03], addr & 0xff);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	writeb_delayed(0x04, priv->i2c_mem + 0x00);
	//ACCESS_ONCE(ptr[align_mem_offset + 0x00], 0x04);

	/* check not busy */
	count = 100;
	while (!(readb(priv->i2c_mem + 0x01) & 0x01) && count) {
		count -= 1;
		udelay(100);
	}

	/* check error */
	if (readb(priv->i2c_mem + 0x01) & (BIT(1) | BIT(2))) {
#if 0		
		if (readb(priv->i2c_mem + 0x01) & (BIT(1)))
			pr_info("%s: Read I2C NACK\n", __func__);

		if (readb(priv->i2c_mem + 0x01) & (BIT(2)))
			pr_info("%s: Read I2C TIMEOUT\n", __func__);
#endif
		return -ETIMEDOUT;
	} else {
		/* Get Data */
		*data = readb(priv->i2c_mem + 0x04) & 0xff;
	}

	return 0;
}

static int __WriteI2CData_F81504_CONF__(struct pci_dev *pdev, unsigned int dev,
					unsigned int addr, unsigned int data)
{
	int delay = 300;
	int status = 0;
	int count;
	unsigned char tmp;

#if 0
	/* Reset */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, 0x80);

	count = 1000;
	while (--count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, &tmp);
		if (!tmp)
			break;

		udelay(delay);
	}

	if (!count) {
		pr_info("%s: timeout: %d\n", __func__, __LINE__);
		return -ETIMEDOUT;
	}
#endif

	/* device address */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x02 * 4, dev & 0xff);

	/* enable WR_CMD_EN */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x06 * 4, 0x03);

	/* clear busy flag */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, 0xff);

	/* Set Reg addr */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x03 * 4, addr & 0xff);

	/* Set data */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x04 * 4, data & 0xff);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x00 * 4, 0x02);

	/* check not busy */
	//ptr[align_mem_offset + 0x01]

	//pci_read_config_byte(int addr,unsigned char * data)
	count = 1000;
	while (--count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, &tmp);
		if (tmp & 0x01)
			break;

		udelay(delay);
	}

	/* check error */
	if (tmp & (BIT(1) | BIT(2))) {
#if 0		
		if (tmp & (BIT(1)))
			fprintf(stderr, "%s: Read I2C NACK\n", __func__);

		if (tmp & (BIT(2)))
			fprintf(stderr, "%s: Read I2C TIMEOUT\n", __func__);
#endif
		pr_info("%s: timeout: %d\n", __func__, __LINE__);
		status = -ETIMEDOUT;
	}

	return status;
}

static int __ReadI2CData_F81504_CONF__(struct pci_dev *pdev, unsigned int dev,
				       unsigned int addr, unsigned int *data)
{
	int delay = 300;
	int status = 0;
	int count;
	unsigned char tmp;

#if 0
	/* Reset */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, 0x80);

	count = 1000;
	while (--count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x07 * 4, &tmp);
		if (!tmp)
			break;

		udelay(delay);
	}

	if (!count) {
		pr_info("%s: timeout: %d\n", __func__, __LINE__);
		return -ETIMEDOUT;
	}
#endif

	/* device address */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x02 * 4, dev & 0xff);
	//msleep(delay);

	/* enable WR_CMD_EN */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x06 * 4, 0x03);
	//msleep(delay);

	/* clear busy flag */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, 0xff);
	//msleep(delay);

	/* Set Reg addr */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x03 * 4, addr & 0xff);
	//msleep(delay);

	/* Start Read, 0x04: EEPROM Read, 0x10: DBI Read?? */
	pci_write_config_byte(pdev, F81504_I2C_CONF + 0x00 * 4, 0x04);
	//msleep(delay);

	/* check not busy */
	//ptr[align_mem_offset + 0x01]

	//pci_read_config_byte(int addr,unsigned char * data)
	count = 1000;
	while (--count) {
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x01 * 4, &tmp);
		if (tmp & 0x01)
			break;

		udelay(delay);
	}

	/* check error */
	if (tmp & (BIT(1) | BIT(2))) {
#if 0		
		if (tmp & (BIT(1)))
			fprintf(stderr, "%s: Read I2C NACK\n", __func__);

		if (tmp & (BIT(2)))
			fprintf(stderr, "%s: Read I2C TIMEOUT\n", __func__);
#endif
		pr_info("%s: timeout: %d\n", __func__, __LINE__);
		status = -ETIMEDOUT;
	} else {
		/* Get Data */
		pci_read_config_byte(pdev, F81504_I2C_CONF + 0x04 * 4, &tmp);
		*data = tmp & 0xff;
	}

	//pr_info("%s: tmp: %x\n", __func__, tmp);

	return status;
}

static int __WriteI2CData_F81504__(struct pci_dev *pdev, unsigned int dev,
				   unsigned int addr, unsigned int data)
{
	struct f81504a_priv *priv = pci_get_drvdata(pdev);

	if (priv->is_i2c_mem)
		return __WriteI2CData_F81504_MEM__(pdev, dev, addr, data);
	else
		return __WriteI2CData_F81504_CONF__(pdev, dev, addr, data);
}

static int __ReadI2CData_F81504__(struct pci_dev *pdev, unsigned int dev,
				  unsigned int addr, unsigned int *data)
{
	struct f81504a_priv *priv = pci_get_drvdata(pdev);

	if (priv->is_i2c_mem)
		return __ReadI2CData_F81504_MEM__(pdev, dev, addr, data);
	else
		return __ReadI2CData_F81504_CONF__(pdev, dev, addr, data);
}

static int __WriteI2CData_Mask_F81504__(struct pci_dev *pdev, unsigned int dev,
					unsigned int addr, unsigned int mask,
					unsigned int data)
{
	unsigned int tmp;
	int status;

	status = __ReadI2CData_F81504__(pdev, dev, addr, &tmp);
	if (status)
		return status;

	tmp = (tmp & ~mask) | (data & mask);

	return __WriteI2CData_F81504__(pdev, dev, addr, tmp);
}

static void f81504_dump_f75111(struct pci_dev *dev)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	u8 dump_tbl[] = { 0x10, 0x11, 0x12, 0x20, 0x21, 0x22 };
	int i, r, tmp;

	for (i = 0; i < ARRAY_SIZE(dump_tbl); ++i) {
		r = __ReadI2CData_F81504__(dev, priv->gpio_dev_id, dump_tbl[i],
					   &tmp);
		pr_info("%s: reg: %02x, data: %02x\n", __func__, dump_tbl[i],
			tmp);
	}
}

static int get_gpio_switch_devid(struct pci_dev *pdev, int *devid)
{
	unsigned int tmp;
	int status;

	*devid = 0x9c;
	status = __ReadI2CData_F81504__(pdev, *devid, 0x5d, &tmp);
	if (status) {
		*devid = 0x6e;

		status = __ReadI2CData_F81504__(pdev, *devid, 0x5d, &tmp);
		if (status) {
			dev_info(
				&pdev->dev,
				"Cant found any GPIO IC, no set_gpio support (1)\n");
			return -ENODEV;
		}
	}

	//pr_info("%s: tmp: %x\n", __func__, tmp);

	if (tmp != 0x19) {
		dev_info(&pdev->dev,
			 "Cant found any GPIO IC, no set_gpio support (2)\n");
		return -ENODEV;
	}

	dev_info(&pdev->dev, "found GPIO IC(%02x), set_gpio supported\n",
		 *devid);

	return 0;
}

static int get_transceiver_type(struct pci_dev *pdev, int dev_id, int idx,
				transceiver_type *type)
{
	unsigned int tmp;
	int status;

	/* read F75111 GPIO 17 for all option: 433/439 or 435 */
	status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10, BIT(7), 0);
	status = __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
	if (status) {
		pr_info("%s: Read all transceiver type failed\n", __func__);
		return status;
	}

	if (tmp & BIT(7)) {
		status = __ReadI2CData_F81504__(pdev, dev_id, 0x42, &tmp);
		if (status) {
			pr_info("%s: Read one transceiver type failed\n",
				__func__);
			return status;
		}

		if (tmp & BIT(idx))
			*type = transceiver_type_f81439;
		else
			*type = transceiver_type_f81433;

	} else {
		*type = transceiver_type_f81435;
	}

	return 0;
}

static int is_hw_fix_mode(struct pci_dev *pdev, int dev_id, int idx,
			  int *is_fix, int *cur_mode)
{
	int status, tmp, mode;

	switch (idx) {
	case 0:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 0, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
		mode = (tmp >> 0) & 0x07;
		break;
	case 1:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 4, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x12, &tmp);
		mode = (tmp >> 4) & 0x07;
		break;
	case 2:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 0, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x22, &tmp);
		mode = (tmp >> 0) & 0x07;
		break;
	case 3:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 4, 0);
		status |= __ReadI2CData_F81504__(pdev, dev_id, 0x22, &tmp);
		mode = (tmp >> 4) & 0x07;
		break;
	default:
		return -EINVAL;
	}

	if (mode != 0x07)
		*is_fix = 1;
	else
		*is_fix = 0;

	if (cur_mode)
		*cur_mode = mode;

	return status;
}

static int change_gpio(struct pci_dev *pdev, int dev_id, int idx, int mode)
{
	unsigned int tmp = 0;
	int status;

	if (mode & BIT(0))
		tmp |= BIT(2);
	if (mode & BIT(1))
		tmp |= BIT(1);
	if (mode & BIT(2))
		tmp |= BIT(0);

	switch (idx) {
	case 0:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 0, 0x07 << 0);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x11,
						       0x07 << 0, tmp << 0);
		break;
	case 1:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x10,
						      0x07 << 4, 0x07 << 4);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x11,
						       0x07 << 4, tmp << 4);
		break;
	case 2:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 0, 0x07 << 0);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x21,
						       0x07 << 0, tmp << 0);
		break;
	case 3:
		status = __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x20,
						      0x07 << 4, 0x07 << 4);
		status |= __WriteI2CData_Mask_F81504__(pdev, dev_id, 0x21,
						       0x07 << 4, tmp << 4);
		break;
	default:
		status = -1;
		break;
	}

	return status;
}

static ssize_t gpio_ctrl_store(struct device *dev,
			       struct device_attribute *attr, const char *buf,
			       size_t count)
{
	struct f81504a_priv *priv;
	struct pci_dev *pdev;
	int i, r;
	unsigned int data[2];
	u8 *p;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	for (i = 0; i < 2; ++i) {
		p = strsep((char **)&buf, " ");
		if (!p) {
			return -EINVAL;
		}

		r = kstrtoint(p, 0, &data[i]);
		if (r) {
			return r;
		}

		if (data[i] >= 256 || data[i] < 0) {
			return -EINVAL;
		}
	}

	if (data[0] >= 4)
		return -EINVAL;

	if (data[1] >= 8)
		return -EINVAL;

	mutex_lock(&priv->locker);

	if (priv->gpio_mode[data[0]] == USER_FIX_MODE) {
		pr_info("%s: cant change hw_fix port: %d\n", __func__,
			data[0]);
		r = -EPERM;
		goto end;
	}

	r = check_transceiver_mode(priv->transceiver_type[data[0]], data[1]);
	if (r)
		goto end;

	priv->gpio_mode[data[0]] = data[1];

	//pr_info("%s: idx: %d, data: %d\n", __func__, data[0], data[1]);
	r = change_gpio(pdev, priv->gpio_dev_id, data[0], data[1]);
end:
	mutex_unlock(&priv->locker);

	//f81504_dump_f75111(pdev);

	if (!r)
		return count;

	pr_info("%s: r: %d\n", __func__, r);
	return r;
}

static ssize_t gpio_ctrl_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct f81504a_priv *priv;
	struct pci_dev *pdev;
	int i, len = 0;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	mutex_lock(&priv->locker);

	for (i = 0; i < 4; ++i)
		len += sprintf(&buf[len], "%d \n", priv->gpio_mode[i]);

	mutex_unlock(&priv->locker);

	return len;
}

static DEVICE_ATTR_RW(gpio_ctrl);

static ssize_t transceiver_show(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct f81504a_priv *priv;
	struct pci_dev *pdev;
	int i, len = 0;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	for (i = 0; i < 4; ++i)
		len += sprintf(&buf[len], "%d\n", priv->transceiver_type[i]);

	return len;
}

static DEVICE_ATTR_RO(transceiver);

static ssize_t f75111_show(struct device *dev, struct device_attribute *attr,
			   char *buf)
{
	struct f81504a_priv *priv;
	struct pci_dev *pdev;

	pdev = container_of(dev, struct pci_dev, dev);
	priv = pci_get_drvdata(pdev);

	f81504_dump_f75111(pdev);

	return 0;
}

static DEVICE_ATTR_RO(f75111);

static int f81504_init_f75111_info(struct pci_dev *dev, bool is_first)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	uint32_t i2c_mem, i;
	int r, is_fix;
	u8 tmp;

	mutex_lock(&priv->locker);

	if ((dev->device >> 8) == 0x13) {
		priv->is_i2c_mem = false;
	} else {
		pci_read_config_byte(dev, 0xb9, &tmp);
		tmp |= 0xe0;
		pci_write_config_byte(dev, 0xb9, tmp);

		tmp = 0xa0;
		pci_write_config_byte(dev, F81504_I2C_CONF + 0x02 * 4, tmp);

		priv->is_i2c_mem = false;
		for (i = 0; i < 4; ++i) {
			pci_read_config_byte(
				dev, F81504_I2C_CONF + i + 0x02 * 4, &tmp);
			if (tmp != 0xa0) {
				priv->is_i2c_mem = true;
				break;
			}
		}
	}

	if (priv->is_i2c_mem) {
		pci_read_config_dword(dev, 0x14, &i2c_mem);
		pci_write_config_dword(dev, 0xb4, i2c_mem);

		priv->i2c_mem = pci_ioremap_bar(dev, 1);
	}

	r = get_gpio_switch_devid(dev, &priv->gpio_dev_id);
	if (r) {
		priv->gpio_dev_id = NO_GPIO_DEV;
		//pr_err("%s: get_gpio_switch_devid err: %d\n", __func__, r);
		goto end;
	}

#if 0
	unsigned int data[2];

	//test
	pr_info("%s: dev_id: %x\n", __func__, priv->gpio_dev_id);

	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x15, 0xff);
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x25, 0xff);
	
	//f81504_dump_f75111(dev);
	udelay(100);

	__ReadI2CData_F81504__(dev, priv->gpio_dev_id, 0x12, &data[0]);
	__ReadI2CData_F81504__(dev, priv->gpio_dev_id, 0x22, &data[1]);

	pr_info("%s: %x %x\n", __func__, data[0], data[1]);

	pr_info("%s: M0_1: %d\n", __func__, !!(data[0] & BIT(0)));
	pr_info("%s: M1_1: %d\n", __func__, !!(data[0] & BIT(1)));
	pr_info("%s: M2_1: %d\n", __func__, !!(data[0] & BIT(2)));
	
	pr_info("%s: M0_2: %d\n", __func__, !!(data[0] & BIT(4)));
	pr_info("%s: M1_2: %d\n", __func__, !!(data[0] & BIT(5)));
	pr_info("%s: M2_2: %d\n", __func__, !!(data[0] & BIT(6)));

	pr_info("%s: M0_3: %d\n", __func__, !!(data[1] & BIT(0)));
	pr_info("%s: M1_3: %d\n", __func__, !!(data[1] & BIT(1)));
	pr_info("%s: M2_3: %d\n", __func__, !!(data[1] & BIT(2)));

	pr_info("%s: M0_4: %d\n", __func__, !!(data[1] & BIT(4)));
	pr_info("%s: M1_4: %d\n", __func__, !!(data[1] & BIT(5)));
	pr_info("%s: M2_4: %d\n", __func__, !!(data[1] & BIT(6)));
	
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x15, 0x00);
	__WriteI2CData_F81504__(dev, priv->gpio_dev_id, 0x25, 0x00);

	goto end;
#endif

	for (i = 0; i < 4; ++i) {
		r = get_transceiver_type(dev, priv->gpio_dev_id, i,
					 &priv->transceiver_type[i]);
		if (r) {
			pr_err("%s: get_transceiver_type err: %d\n", __func__,
			       r);
			goto end;
		}

		r = is_hw_fix_mode(dev, priv->gpio_dev_id, i, &is_fix, NULL);
		if (r) {
			pr_err("%s: is_hw_fix_mode err: %d\n", __func__, r);
			goto end;
		}

		if (is_first)
			priv->gpio_mode[i] = 1;

		if (!is_fix) {
			r = change_gpio(dev, priv->gpio_dev_id, i,
					priv->gpio_mode[i]);
			if (r) {
				pr_err("%s: change_gpio err: %d\n", __func__,
				       r);
				goto end;
			}
		} else {
			priv->gpio_mode[i] = USER_FIX_MODE;
		}
	}

end:
	mutex_unlock(&priv->locker);
	return r;
}

static int f81504_prepare_mode_pin(struct pci_dev *dev)
{
	int r;

	r = f81504_init_f75111_info(dev, true);
	if (r)
		return r;

	r = device_create_file(&dev->dev, &dev_attr_f75111);
	if (r)
		goto attr_f75111_err;

	r = device_create_file(&dev->dev, &dev_attr_gpio_ctrl);
	if (r)
		goto attr_gpio_ctrl_err;

	r = device_create_file(&dev->dev, &dev_attr_transceiver);
	if (r)
		goto attr_transceiver_err;

	return 0;

attr_transceiver_err:
	device_remove_file(&dev->dev, &dev_attr_gpio_ctrl);
attr_gpio_ctrl_err:
	device_remove_file(&dev->dev, &dev_attr_f75111);
attr_f75111_err:

	return r;
}

static void f81504_remove_mode_pin(struct pci_dev *dev)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);

	if (priv->gpio_dev_id != NO_GPIO_DEV) {
		device_remove_file(&dev->dev, &dev_attr_f75111);
		device_remove_file(&dev->dev, &dev_attr_transceiver);
		device_remove_file(&dev->dev, &dev_attr_gpio_ctrl);
	}

	if (priv->i2c_mem)
		iounmap(priv->i2c_mem);
}

static int f81504a_reapply_gpio_mode_setting(struct pci_dev *pci_dev,
					     u8 port_idx)
{
	struct f81504a_priv *priv = pci_get_drvdata(pci_dev);
	struct uart_8250_port *up;
	bool is_rs485, is_invert;
	int r, mode, is_fix = 0;
	transceiver_type type;

	if (priv->gpio_dev_id == NO_GPIO_DEV)
		return 0;

	if (port_idx >= 4)
		return 0;

	if (priv->line_data[port_idx] < 0) {
		pr_warn("%s: line data err: %d, %d\n", __func__, port_idx,
			priv->line_data[port_idx]);
		return -EINVAL;
	}

	up = serial8250_get_port(priv->line_data[port_idx]);
	r = is_hw_fix_mode(pci_dev, priv->gpio_dev_id, port_idx, &is_fix,
			   &mode);
	if (r)
		return r;

	is_rs485 = is_invert = false;

	if (is_fix) {
		priv->gpio_mode[port_idx] = USER_FIX_MODE;
		type = priv->transceiver_type[port_idx];

		switch (type) {
		case transceiver_type_f81439:
		case transceiver_type_f81435:
		case transceiver_type_f81433:
			//pr_info("%s: type: %d, mode: %d\n", __func__, type,
			//	mode);

			if (m_n_rs485_Array[type][mode] != 0) {
				is_rs485 = !!(m_n_rs485_Array[type][mode] &
					      BIT(0));
				is_invert = !!(m_n_rs485_Array[type][mode] &
					       BIT(1));
			}
			break;

		case transceiver_type_invalid:
		default:
			pr_warn("%s: transceiver type unknown: %d\n", __func__,
				type);
			return -EINVAL;
		}

		if (is_rs485) {
			up->port.rs485.flags = SER_RS485_ENABLED;

			if (is_invert)
				up->port.rs485.flags = SER_RS485_RTS_ON_SEND;
			else
				up->port.rs485.flags =
					SER_RS485_RTS_AFTER_SEND;

		} else {
			up->port.rs485.flags = SER_RS485_RTS_AFTER_SEND;
		}
	} else {
		if (priv->gpio_mode[port_idx] == USER_FIX_MODE && mode == 7)
			priv->gpio_mode[port_idx] = 1;

		r = change_gpio(pci_dev, priv->gpio_dev_id, port_idx,
				priv->gpio_mode[port_idx]);
		if (r)
			return r;
	}

	pr_debug(
		"%s: gpio_dev_id: %x, is_fix: %d, read mode: %xh, cur mode: %xh\n",
		__func__, priv->gpio_dev_id, is_fix, mode,
		priv->gpio_mode[port_idx]);

	return 0;
}

#else
static int f81504_init_f75111_info(struct pci_dev *dev, , bool is_first)
{
	return 0;
}

static int f81504_prepare_mode_pin(struct pci_dev *dev)
{
	return 0;
}

static void f81504_remove_mode_pin(struct pci_dev *dev)
{
}

static int f81504a_reapply_gpio_mode_setting(struct pci_dev *pci_dev,
					     u8 port_idx)
{
	return 0;
}
#endif

static int f81504a_find_clk(speed_t baudrate)
{
	int idx;

	for (idx = 0; idx < ARRAY_SIZE(baudrate_table); ++idx) {
		if (baudrate <= baudrate_table[idx] &&
		    baudrate_table[idx] % baudrate == 0)
			return idx;
	}

	return -EINVAL;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 1, 0)
static void f81504a_set_termios(struct uart_port *port,
				struct ktermios *termios, struct ktermios *old)
#else
static void f81504a_set_termios(struct uart_port *port,
				struct ktermios *termios,
				const struct ktermios *old)
#endif
{
	struct pci_dev *pdev = to_pci_dev(port->dev);
	speed_t baudrate, old_baudrate;
	speed_t baud_list[] = { 0, F81504A_DEF_BAUDRATE,
				F81504A_DEF_BAUDRATE };
	u8 port_idx = *(u8 *)port->private_data;
	int baud_idx = 0;
	u8 tmp;
	int i;

	baudrate = old_baudrate = tty_termios_baud_rate(termios);
	baud_list[0] = baudrate;

	dev_dbg(&pdev->dev, "%s: baudrate: %d\n", __func__, baudrate);

	if (old)
		baud_list[1] = tty_termios_baud_rate(old);

	if (baudrate) {
		for (i = 0; i < ARRAY_SIZE(baud_list); ++i) {
			baud_idx = f81504a_find_clk(baud_list[i]);
			if (baud_idx >= 0) {
				dev_dbg(&pdev->dev,
					"%s: baud_list: %d, baud_idx: %d\n",
					__func__, baud_list[i], baud_idx);

				baudrate = baud_list[i];
				break;
			}
		}

		if (old_baudrate < 115200 && old_baudrate != baudrate) {
			dev_dbg(&pdev->dev,
				"force change to baud clk 115200\n");

			baudrate = old_baudrate;
			baud_idx = 0;
		}

		tty_termios_encode_baud_rate(termios, baudrate, baudrate);
	}

	if (baud_idx < 0 || baud_idx >= ARRAY_SIZE(clock_table)) {
		dev_err(&pdev->dev, "Can't found suitable clock source\n");
		return;
	}

	pci_read_config_byte(pdev, F81504A_UART_CTRL_REG(port_idx), &tmp);

	if (force_high_speed) {
		baud_idx = ARRAY_SIZE(clock_table) - 1;
		tty_termios_encode_baud_rate(termios, 6000000, 6000000);
	}

	dev_dbg(&pdev->dev, "%s: baudrate idx: %d, reg: %xh\n", __func__,
		baud_idx, clock_table[baud_idx]);

	tmp = (tmp & ~F81504A_CLK_MASK) | clock_table[baud_idx] |
	      F81504A_UART_EN;

	port->uartclk = baudrate_table[baud_idx] * 16;

	pci_write_config_byte(pdev, F81504A_UART_CTRL_REG(port_idx), tmp);

#if F81504A_INTERNAL_LOOP
	struct uart_8250_port *up =
		container_of(port, struct uart_8250_port, port);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 17, 0)
	//if (port->line != 9)
	up->mcr_force |= UART_MCR_LOOP;
	//printk(KERN_INFO "%s: up->mcr: %x\n", __func__, up->mcr);
	//printk(KERN_INFO "%s: up->mcr_force: %x\n", __func__, up->mcr_force);
	//serial8250_do_set_mctrl(port, TIOCM_LOOP);
#else
	serial8250_do_set_mctrl(port, TIOCM_LOOP);
#endif
#endif

	serial8250_do_set_termios(port, termios, old);
}

static int f81504a_fill_port_register(struct pci_dev *dev, int idx)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	struct serial_rs485 *rs485 = NULL;
	u8 tmp = 0;

	/* Enable UART */
	pci_write_config_byte(dev, F81504A_UART_CTRL_REG(idx),
			      F81504A_UART_EN);

	/* Select 128-byte FIFO and 8x FIFO threshold */
	pci_write_config_byte(dev, F81504A_UART_FIFO_REG(idx),
			      F81504A_TRIGGER_8X_MODE | F81504A_FIFO_256_MODE);

	rs485 = &priv->rs485[idx];
	pci_read_config_byte(dev, F81504A_UART_IRQ_SHARE_REG(idx), &tmp);

	tmp &= ~(F81504A_UART_RS485_INV | F81504A_UART_RS485_ENABLE);

	//pr_info("%s: 111 tmp: %02x\n", __func__, tmp);
	//pr_info("%s: rs485->flags: %x\n", __func__, rs485->flags);

	if (rs485->flags & SER_RS485_ENABLED) {
		//pr_info("%s: 333\n", __func__);
		tmp |= F81504A_UART_RS485_ENABLE;

		if (!(rs485->flags & SER_RS485_RTS_ON_SEND)) {
			//pr_info("%s: 444\n", __func__);
			tmp |= F81504A_UART_RS485_INV;
		}
	}

	//pr_info("%s: 222 tmp: %02x\n", __func__, tmp);

	//tmp |= F81504A_UART_IRQ_SHARE;
	//tmp = 0;

	pci_write_config_byte(dev, F81504A_UART_IRQ_SHARE_REG(idx), tmp);

	return 0;
}

static int __f81504a_rs485_config(struct uart_port *port,
				  struct serial_rs485 *rs485)
{
	struct pci_dev *pdev = to_pci_dev(port->dev);
	struct f81504a_priv *priv = pci_get_drvdata(pdev);
	u8 idx = *(u8 *)port->private_data;

	if (idx < 0)
		return -EINVAL;

	priv->rs485[idx] = *rs485;
	//pr_info("%s: rs485->flags: %x\n", __func__, rs485->flags);
	f81504a_fill_port_register(pdev, idx);

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
static int f81504a_rs485_config(struct uart_port *port,
				struct ktermios *termios,
				struct serial_rs485 *rs485)
{
	return __f81504a_rs485_config(port, rs485);
}

#else
static int f81504a_rs485_config(struct uart_port *port,
				struct serial_rs485 *rs485)
{
	return __f81504a_rs485_config(port, rs485);
}
#endif

static unsigned int f81504a_io_serial_in(struct uart_port *p, int offset)
{
	offset = offset << p->regshift;

	return inb(p->iobase + offset);
}

static void f81504a_io_serial_out(struct uart_port *p, int offset, int value)
{
	offset = offset << p->regshift;
	outb(value, p->iobase + offset);
}

static unsigned int f81504a_mem_serial_in(struct uart_port *p, int offset)
{
	unsigned char __iomem *mem = p->membase;

	offset = offset << p->regshift;
	return readb(mem + offset);
}

static void f81504a_mem_serial_out(struct uart_port *p, int offset, int value)
{
	unsigned char __iomem *mem = p->membase;

	offset = offset << p->regshift;
	writeb(value, mem + offset);
}

#if 0
static bool f81504a_check_need_fixup(struct pci_dev *dev)
{
	bool need_fixup = false;
	int count = 10;
	u8 tmp;

	/* use UART1 to test PCIe configure Byte Enable */
	pci_write_config_byte(dev, 0x209, 0x20);
	pci_write_config_byte(dev, 0x2a0, 0x01);

	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 0, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 2, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 5, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 6, &tmp);

	pci_write_config_byte(dev, F81504A_UART_CONF(0) + 1, 0x01);
	pci_write_config_byte(dev, F81504A_UART_CONF(0) + 2, 0xc7);
	//pci_write_config_byte(dev, F81504A_UART_CONF(0) + 4, 0x18);
	pci_write_config_byte(dev, F81504A_UART_CONF(0) + 4,
			      0x10); /* interrupt will not send out */
	pci_write_config_byte(dev, F81504A_UART_CONF(0) + 0, 0xaa);

	while (--count) {
		udelay(100);

		pci_read_config_byte(dev, F81504A_UART_CONF(0) + 2, &tmp);
		if (tmp == 0xcc)
			break;
	}

	if (tmp != 0xcc)
		need_fixup = true;

	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 0, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 2, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 5, &tmp);
	pci_read_config_byte(dev, F81504A_UART_CONF(0) + 6, &tmp);

	dev_info(&dev->dev, "need_fixup: %x\n", need_fixup);

	return need_fixup;
}
#endif

static irqreturn_t f81504a_dummy_interrupt(int irq, void *data)
{
	return IRQ_RETVAL(0);
}

#define UART_CAP_FIFO (1 << 8) /* UART has FIFO */

struct uart_8250_port *serial8250_get_port(int line);
//static int f81504a_uart_handle_irq(struct uart_port *port);

static int f81504a_8250_startup(struct uart_port *port)
{
	struct pci_dev *pci_dev = container_of(port->dev, struct pci_dev, dev);
	u8 port_idx = *(u8 *)port->private_data;
	int r;

	r = f81504a_reapply_gpio_mode_setting(pci_dev, port_idx);
	if (r)
		return r;

	r = __f81504a_rs485_config(port, &port->rs485);
	if (r)
		return r;

	return serial8250_do_startup(port);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
static const struct serial_rs485 fintek_8250_rs485_supported = {
	.flags = SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND |
		 SER_RS485_RTS_AFTER_SEND,
};
#endif

static void f81504a_conf_init(struct pci_dev *dev)
{
	u8 tmp;

#if 0
	switch (mmio_mode) {
	case 0:
#if EXTRA_GPIO_CHANGE_MODE
		pci_write_config_byte(dev, 0x209, 0xe0);
#else
		pci_write_config_byte(dev, 0x209, 0x40);
#endif
		break;

	case 1:
		pci_write_config_byte(dev, 0x209, 0x80);
		break;

	case 2:
		pci_write_config_byte(dev, 0x209, 0xe0);
		break;
	}
#endif

	pci_read_config_byte(dev, 0x209, &tmp);
	tmp |= 0xe0;

#if !F81504A_CLK_DIV_2
	tmp |= BIT(4); // force no div2
#else
	tmp &= ~BIT(4);
#endif

	pci_write_config_byte(dev, 0x209, tmp);
}

static int f81504a_register_port(struct pci_dev *dev, int offset, int irq,
				 bool fixup)
{
	//struct f81504a_priv *priv = pci_get_drvdata(dev);
	struct uart_8250_port port;
	u8 *data;
	int line;

	memset(&port, 0, sizeof(port));

	switch (mmio_mode) {
	case 0:
		dev_dbg(&dev->dev, "mmio_mode: mem\n");

		port.port.iotype = UPIO_MEM;
		port.port.flags |= UPF_IOREMAP | UPF_SHARE_IRQ;
		port.port.mapbase = pci_resource_start(dev, 0) + 8 * offset;
		port.port.serial_in = f81504a_mem_serial_in;
		port.port.serial_out = f81504a_mem_serial_out;
		break;

	case 1:
		dev_dbg(&dev->dev, "mmio_mode: io\n");

		port.port.iotype = UPIO_PORT;
		port.port.iobase = pci_resource_start(dev, 1) + 8 * offset;
		port.port.serial_in = f81504a_io_serial_in;
		port.port.serial_out = f81504a_io_serial_out;
		break;

	case 2:
		port.port.iotype = UPIO_MEM;
		port.port.flags |= UPF_IOREMAP;
		port.port.mapbase = pci_resource_start(dev, 0) + 8 * offset;

		if (fixup) {
			port.port.serial_in = f81504a_mem_serial_in;
			dev_dbg(&dev->dev, "mmio_mode: mem in/ conf out\n");
		} else {
			port.port.serial_in = f81504a_conf_serial_in;
			dev_dbg(&dev->dev, "mmio_mode: conf\n");
		}

		port.port.serial_out = f81504a_conf_serial_out;
		break;
	}

	data = devm_kzalloc(&dev->dev, sizeof(u8), GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	*data = (u8)offset;

	port.port.irq = irq;
	//port.port.flags |= UPF_SKIP_TEST | UPF_FIXED_TYPE;
	port.port.flags |= UPF_SKIP_TEST | UPF_FIXED_TYPE;
	port.port.flags |= UPF_SHARE_IRQ | UPF_BOOT_AUTOCONF;

	// non-usage
	if (pci_dev_msi_enabled(dev))
		port.port.irqflags |= IRQF_NO_SUSPEND;

	port.port.startup = f81504a_8250_startup;
	port.port.uartclk = baudrate_table[0] * 16; //1843200;
	port.port.dev = &dev->dev;
	port.port.type = PORT_16550A;
	port.port.fifosize = 128;
	port.port.private_data = data; /* save current idx */
	port.tx_loadsz = 32;

	port.port.set_termios = f81504a_set_termios;
	port.port.rs485_config = f81504a_rs485_config;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)
	memcpy(&port.port.rs485_supported, &fintek_8250_rs485_supported,
	       sizeof(fintek_8250_rs485_supported));
#endif

	line = serial8250_register_8250_port(&port);

	return line;
}

static int f81504a_probe(struct pci_dev *dev,
			 const struct pci_device_id *dev_id)
{
	struct f81504a_priv *priv;
	int status;
	int i, r;
	int irq_count = 1;
	int irq;
	bool need_fixup = 0;

	dev_info(&dev->dev, "%s: version: %s\n", __func__, VERSION);

	if (force_high_speed && !F81504A_CLK_DIV_2) {
		pr_info("%s: force_high_speed feature must disable F81504A_CLK_DIV_2\n",
			__func__);
	}

	priv = devm_kzalloc(&dev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	pci_set_drvdata(dev, priv);
	mutex_init(&priv->locker);
	memcpy(&priv->dev_id, dev_id, sizeof(struct pci_device_id));

	f81504a_conf_init(dev);

	r = f81504_prepare_mode_pin(dev);
	if (r && r != -ENODEV)
		return r;

	status = pcim_enable_device(dev);
	if (status)
		return status;

#if 0
	if (pci_dev_run_wake(dev)) {
		dev_info(&dev->dev, "PME cap\n");

		status = pci_enable_wake(dev, PCI_D0, 1);
		if (status)
			dev_info(&dev->dev,
				 "PME PCI_D3hot enable failed: %x\n", status);

		pci_pme_active(dev, true);
	}
#endif

#if 1
	pci_write_config_byte(dev, 0x20c, 0xb);
	pci_write_config_byte(dev, 0x20d, 0x7);
	pci_write_config_byte(dev, 0x20e, 0xb);
#else
	pci_write_config_byte(dev, 0x20c, 0xb);
	pci_write_config_byte(dev, 0x20d, 0x7);
	pci_write_config_byte(dev, 0x20e, 0xb);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
	if (en_msi) {
		irq_count = pci_alloc_irq_vectors(
			dev, 1, max_msi, PCI_IRQ_ALL_TYPES | PCI_IRQ_AFFINITY);
	} else {
		irq_count = pci_alloc_irq_vectors(
			dev, 1, max_msi, PCI_IRQ_LEGACY | PCI_IRQ_AFFINITY);
	}
	if (irq_count < 0)
		return irq_count;

	dev_info(&dev->dev, "We had %d MSI allocated\n", irq_count);
#else
	if (en_msi)
		irq_count = pci_enable_msi_range(dev, 1, max_msi);
	else
		irq_count = -1;
	if (irq_count < 0) {
		irq_count = 1;
		pci_intx(dev, 1);
	}
#endif

	priv->irq_cnt = irq_count;

	if (irq_count > 1) {
		// install dummy MSI interrupt

		for (i = 0; i < priv->irq_cnt; ++i) {
			status = devm_request_irq(
				&dev->dev, dev->irq + i,
				f81504a_dummy_interrupt,
				IRQF_SHARED | IRQF_NO_SUSPEND, "fintek", dev);
			if (status) {
				pr_err("%s: register interrupt failed: %d\n",
				       __func__, status);
				return status;
			}
		}
	}

	dev_info(&dev->dev, "IRQ start from: %d\n", dev->irq);

	//need_fixup = f81504a_check_need_fixup(dev);

	for (i = 0; i < dev_id->driver_data; ++i)
		f81504a_fill_port_register(dev, i);

	for (i = 0; i < dev_id->driver_data; ++i) {
		irq = dev->irq + ((i + 1) % irq_count);
		//pr_info("%s: i: %d, irq: %d\n", __func__, i, irq);

		priv->line_data[i] =
			f81504a_register_port(dev, i, irq, need_fixup);
		//pr_info("%s: %d\n", __func__, priv->line_data[i]);
		if (priv->line_data[i] < 0) {
			dev_warn(&dev->dev, "register %d port failed: %d\n", i,
				 priv->line_data[i]);
		}
	}

	device_set_wakeup_enable(&dev->dev, true);
	return 0;
}

static void f81504a_remove(struct pci_dev *dev)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	int i;

	for (i = 0; i < priv->dev_id.driver_data; ++i) {
		if (priv->line_data[i] >= 0)
			serial8250_unregister_port(priv->line_data[i]);
	}

	device_set_wakeup_enable(&dev->dev, false);
	f81504_remove_mode_pin(dev);
}

static int f81504a_suspend(struct pci_dev *dev, pm_message_t state)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	int i;

	if (!priv)
		return -ENODEV;

	for (i = 0; i < priv->dev_id.driver_data; ++i) {
		if (priv->line_data[i] >= 0)
			serial8250_suspend_port(priv->line_data[i]);
	}

	return 0;
}

static int f81504a_resume(struct pci_dev *dev)
{
	struct f81504a_priv *priv = pci_get_drvdata(dev);
	int i;

	if (!priv)
		return -ENODEV;

	f81504a_conf_init(dev);

	for (i = 0; i < priv->dev_id.driver_data; ++i) {
		f81504a_fill_port_register(dev, i);

		if (priv->line_data[i] >= 0)
			serial8250_resume_port(priv->line_data[i]);
	}

	return 0;
}

static const struct pci_device_id f81504a_dev_table[] = {
	/* Fintek PCI serial cards */
	{ PCI_DEVICE(0x1c29, 0x1204), .driver_data = 4 },
	{ PCI_DEVICE(0x1c29, 0x1208), .driver_data = 8 },
	{ PCI_DEVICE(0x1c29, 0x1212), .driver_data = 12 },

	{ PCI_DEVICE(0x1c29, 0x1304), .driver_data = 4 },
	{ PCI_DEVICE(0x1c29, 0x1308), .driver_data = 8 },
	{ PCI_DEVICE(0x1c29, 0x1312), .driver_data = 12 },
	{}
};

MODULE_DEVICE_TABLE(pci, f81504a_dev_table);

static struct pci_driver f81504a_driver = {
	.name = "f81504a_u3-4-8-12",
	.probe = f81504a_probe,
	.remove = f81504a_remove,
#ifdef CONFIG_PM
	.suspend = f81504a_suspend,
	.resume = f81504a_resume,
#endif
	.id_table = f81504a_dev_table,
};

static __init int f81504a_init(void)
{
	spin_lock_init(&global_lock);

	return pci_register_driver(&f81504a_driver);
}

static __exit void f81504a_exit(void)
{
	pci_unregister_driver(&f81504a_driver);
}

module_init(f81504a_init);
module_exit(f81504a_exit);

MODULE_DESCRIPTION("F81504AA/508A/512A U3");
MODULE_AUTHOR("Peter Hong <Peter_Hong@fintek.com.tw>");
MODULE_LICENSE("GPL");
